@@ -27,8 +27,7 @@ struct notify_change_request {
 	struct files_struct *fsp;	/* backpointer for cancel by mid */
 	char request_buf[smb_size];
 	uint32 filter;
-	uint32 max_param_count;
-	uint32 current_bufsize;
+	uint32 max_param;
 	struct notify_mid_map *mid_map;
 	void *backend_data;
 };
@@ -48,19 +47,42 @@ struct notify_mid_map {
 	uint16 mid;
 };
 
+static BOOL notify_change_record_identical(struct notify_change *c1,
+					struct notify_change *c2)
+{
+	/* Note this is deliberately case sensitive. */
+	if (c1->action == c2->action &&
+			strcmp(c1->name, c2->name) == 0) {
+		return True;
+	}
+	return False;
+}
+
 static BOOL notify_marshall_changes(int num_changes,
-				    struct notify_change *changes,
-				    prs_struct *ps)
+				uint32 max_offset,
+				struct notify_change *changes,
+				prs_struct *ps)
 {
 	int i;
 	UNISTR uni_name;
 
+	uni_name.buffer = NULL;
+
 	for (i=0; i<num_changes; i++) {
-		struct notify_change *c = &changes[i];
+		struct notify_change *c;
 		size_t namelen;
 		uint32 u32_tmp;	/* Temp arg to prs_uint32 to avoid
 				 * signed/unsigned issues */
 
+		/* Coalesce any identical records. */
+		while (i+1 < num_changes &&
+			notify_change_record_identical(&changes[i],
+						&changes[i+1])) {
+			i++;
+		}
+
+		c = &changes[i];
+
 		namelen = convert_string_allocate(
 			NULL, CH_UNIX, CH_UTF16LE, c->name, strlen(c->name)+1,
 			&uni_name.buffer, True);
@@ -91,6 +113,11 @@ static BOOL notify_marshall_changes(int num_changes,
 		prs_set_offset(ps, prs_offset(ps)-2);
 
 		SAFE_FREE(uni_name.buffer);
+
+		if (prs_offset(ps) > max_offset) {
+			/* Too much data for client. */
+			return False;
+		}
 	}
 
 	return True;
@@ -126,26 +153,23 @@ static void change_notify_reply_packet(const char *request_buf,
 				    "failed.");
 }
 
-void change_notify_reply(const char *request_buf, uint32 max_param_count,
+void change_notify_reply(const char *request_buf, uint32 max_param,
 			 struct notify_change_buf *notify_buf)
 {
 	char *outbuf = NULL;
 	prs_struct ps;
-	size_t buflen = smb_size+38+max_param_count;
+	size_t buflen;
 
 	if (notify_buf->num_changes == -1) {
 		change_notify_reply_packet(request_buf, NT_STATUS_OK);
+		notify_buf->num_changes = 0;
 		return;
 	}
 
-	if (!prs_init(&ps, 0, NULL, False)
-	    || !notify_marshall_changes(notify_buf->num_changes,
-					notify_buf->changes, &ps)) {
-		change_notify_reply_packet(request_buf, NT_STATUS_NO_MEMORY);
-		goto done;
-	}
+	prs_init(&ps, 0, NULL, MARSHALL);
 
-	if (prs_offset(&ps) > max_param_count) {
+	if (!notify_marshall_changes(notify_buf->num_changes, max_param,
+					notify_buf->changes, &ps)) {
 		/*
 		 * We exceed what the client is willing to accept. Send
 		 * nothing.
@@ -154,6 +178,8 @@ void change_notify_reply(const char *request_buf, uint32 max_param_count,
 		goto done;
 	}
 
+	buflen = smb_size+38+prs_offset(&ps) + 4 /* padding */;
+
 	if (!(outbuf = SMB_MALLOC_ARRAY(char, buflen))) {
 		change_notify_reply_packet(request_buf, NT_STATUS_NO_MEMORY);
 		goto done;
@@ -214,7 +240,7 @@ NTSTATUS change_notify_create(struct files_struct *fsp, uint32 filter,
 	return status;
 }
 
-NTSTATUS change_notify_add_request(const char *inbuf, uint32 max_param_count,
+NTSTATUS change_notify_add_request(const char *inbuf, uint32 max_param,
 				   uint32 filter, BOOL recursive,
 				   struct files_struct *fsp)
 {
@@ -231,12 +257,11 @@ NTSTATUS change_notify_add_request(const char *inbuf, uint32 max_param_count,
 	map->req = request;
 
 	memcpy(request->request_buf, inbuf, sizeof(request->request_buf));
-	request->max_param_count = max_param_count;
-	request->current_bufsize = 0;
+	request->max_param = max_param;
 	request->filter = filter;
 	request->fsp = fsp;
 	request->backend_data = NULL;
-	
+
 	DLIST_ADD_END(fsp->notify->requests, request,
 		      struct notify_change_request *);
 
@@ -408,7 +433,7 @@ static void notify_fsp(files_struct *fsp, uint32 action, const char *name)
 	 */
 
 	change_notify_reply(fsp->notify->requests->request_buf,
-			    fsp->notify->requests->max_param_count,
+			    fsp->notify->requests->max_param,
 			    fsp->notify);
 
 	change_notify_remove_request(fsp->notify->requests);
